(in-package :asn.1)

(defgeneric compile-asn.1 (object &key &allow-other-keys))

(defmethod compile-asn.1 ((file t) &key to &allow-other-keys)
  (let ((result (compile-asn.1 (parse file))))
    (when (pathnamep to)
      (let ((head (merge-pathnames (make-pathname :type "lisp-expr")
                                   to))
            (*package* (find-package :asn.1)))
        (with-open-file (h head :direction :output :if-exists :supersede)
          (format h ";;;; -*- Mode: Lisp -*-~%;;;; Auto-generated from ~A~%" file)
          (pprint (car result) h)
          (format h "~%"))
        (with-open-file (s to :direction :output :if-exists :supersede)
          (format s ";;;; Auto-generated from ~A~%" file)
          (dolist (i (cdr result))
            (pprint i s))
          (format s "~%"))))
    result))

(defmethod compile-asn.1 ((rtl list) &key &allow-other-keys)
  (if (atom rtl)
      rtl
    (compile-asn.1-internal (car rtl) rtl)))

(defgeneric compile-asn.1-internal (type rtl))

(defmethod compile-asn.1-internal ((type symbol) (rtl list))
  `(defunknown ,type))

(defmethod compile-asn.1-internal ((type (eql :module)) (rtl list))
  (declare (ignore type))
  (destructuring-bind (module-name xxx) (cdr rtl)
    (if (null xxx)
        `(()
          (in-package :asn.1))
      (destructuring-bind ((exports imports) body) xxx
        (declare (ignore exports))
        (append `(,(compile-imports imports)
                  (in-package :asn.1)
                  (eval-when (:load-toplevel :execute)
                    (setf *current-module* ',module-name)))
                (mapcar #'compile-asn.1 body))))))

(defun compile-imports (rtl)
  (when rtl
    (let ((imports (cdr rtl)))
      (mapcar #'(lambda (x) (string-downcase (symbol-name (car (last x)))))
              imports))))

(defmethod compile-asn.1-internal ((type (eql :value-assignment)) (rtl list))
  (declare (ignore type))
  (compile-va (cadr rtl) (cdr rtl)))

(defmethod compile-asn.1-internal ((type (eql :define)) (rtl list))
  (declare (ignore type))
  (compile-df (cadr rtl) (cdr rtl)))

(defgeneric compile-va (type rtl))

(defmethod compile-va ((type (eql :object-identifier)) (rtl list))
  (destructuring-bind ((name value)) (cdr rtl)
    (setf value (last value 2))
    (destructuring-bind (parent value) value
      (if (consp parent) (setf parent (car parent)))
      `(defoid ,name (,parent ,value)))))

(defgeneric compile-df (type rtl))

(defmethod compile-df ((type symbol) (rtl list))
  `(defunknown ',type))

#|
(:DEFINE
    OBJECT-IDENTITY
    (|zeroDotZero| ((STATUS |current|)
                    (DESCRIPTION "A value used for null identifiers.")))
    (0 0))
|#

(defmethod compile-df ((type (eql 'OBJECT-IDENTITY)) (rtl list))
  (destructuring-bind ((name options) (parent value)) (cdr rtl)
    (declare (ignore options))
    `(defoid ,name (,(if (and (numberp parent)
                              (zerop parent))
                         '|zero| parent) ,value))))

(defmethod compile-df ((type (eql 'MODULE-IDENTITY)) (rtl list))
  (destructuring-bind ((name options) (parent value)) (cdr rtl)
    (declare (ignore options))
    `(defoid ,name (,parent ,value))))

(defmethod compile-df ((type (eql 'OBJECT-TYPE)) (rtl list))
  (destructuring-bind ((name options) (parent value)) (cdr rtl)
    (declare (ignore options))
    `(defoid ,name (,parent ,value))))

(defmethod compile-df ((type (eql 'OBJECT-GROUP)) (rtl list))
  (destructuring-bind ((name options) (parent value)) (cdr rtl)
    (declare (ignore options))
    `(defoid ,name (,parent ,value))))

(defmethod compile-df ((type (eql 'NOTIFICATION-TYPE)) (rtl list))
  (destructuring-bind ((name options) (parent value)) (cdr rtl)
    (declare (ignore options))
    `(defoid ,name (,parent ,value))))

(defmethod compile-df ((type (eql 'NOTIFICATION-GROUP)) (rtl list))
  (destructuring-bind ((name options) (parent value)) (cdr rtl)
    (declare (ignore options))
    `(defoid ,name (,parent ,value))))
