(in-package :asn.1)

;;; Add a declare to ignore initargs
(defmethod initialize-instance :after ((instance object-id)
                                       &rest initargs &key &allow-other-keys)
  (declare (ignore initargs))
  (if (and (slot-boundp instance 'parent)
           (slot-boundp instance 'name)
           (slot-boundp instance 'value))
      (let ((o (oid-parent instance)))
        (setf (gethash (oid-value instance) (oid-children o)) instance))))

;;; Fix return values, return OBJECT-ID instances now
(defmethod ber-decode-value ((s stream) (type (eql :object-identifier)) length)
  (declare (type stream s)
           (type fixnum length)
           (ignore type))
  (if (zerop length)
      (oid '(0))
    (labels ((get-number (acc len)
               (let* ((byte (read-byte s))
                      (val (logior (ash acc 7) (logand byte 127))))
                 (if (< byte 128) (values val len)
                   (get-number val (1+ len)))))
             (iter (left-length acc head-p)
               (declare (type fixnum left-length)
                        (type list acc))
               (if (zerop left-length)
                   (nreverse acc)
                 (multiple-value-bind (n l) (get-number 0 1)
                   (if head-p
                       (multiple-value-bind (q r) (floor n 40)
                         (iter (- left-length l) (cons r (cons q acc)) nil))
                     (iter (- left-length l) (cons n acc) nil))))))
      (oid (iter length nil t)))))
